"""
Package FuPy - Functional Programming in Python

* FuPy.core loads FuPy.basics, FuPy.tracing, FuPy.laziness, FuPy.fixpoints
* FuPy.all loads FuPy.core and FuPy.prelude

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
